/*
 *  gensio - A library for abstracting stream I/O
 *  Copyright (C) 2018  Corey Minyard <minyard@acm.org>
 *
 *  SPDX-License-Identifier: LGPL-2.1-only
 */

#ifndef GENSIO_NET_H
#define GENSIO_NET_H

enum gensio_tcpd_options {
    GENSIO_TCPD_ON,
    GENSIO_TCPD_PRINT,
    GENSIO_TCPD_OFF,
};

#endif /* GENSIO_NET_H */
